#include "LapAssembler.hpp"

template<>
void LapAssembler::assemble_internal_face_contribution(const SpaMesh::Mesh<InputData::dim>& mesh,
                                                       PETScMatrixType& matrix, 
                                                       PETScVectorType& pRhs, 
				                       FEManagerType& FEManager)
{
  const FEManagerType::ShapeMatrixType& bShapeMatrix             = FEManager.getInternalFaceElemShapeFunction();
  const FEManagerType::ShapeMatrixType& fShapeMatrix             = FEManager.getInternalFaceNeighborShapeFunction();
  const FEManagerType::ShapeDerivMatrixType& bShapeDerivMatrix   = FEManager.getInternalFaceElemShapeFunctionDeriv();     
  const FEManagerType::ShapeDerivMatrixType& fShapeDerivMatrix   = FEManager.getInternalFaceNeighborShapeFunctionDeriv(); 
  const FEManagerType::NormalVectorType& normalVec               = FEManager.getInternalFaceNormalVec();
  const FEManagerType::FaceJacobianVectorType& jacXweightVecSide = FEManager.getInternalFaceJacobianWeightVec();
  const FEManagerType::FaceQpointVectorType& qpXVecSide          = FEManager.getInternalFaceQpointX();
  const LiftingOperatorType::LiftingMatrix& R11 = FEManager.getLifting11(); 
  const LiftingOperatorType::LiftingMatrix& R12 = FEManager.getLifting12(); 
  const LiftingOperatorType::LiftingMatrix& R22 = FEManager.getLifting22(); 
  const LiftingOperatorType::LiftingMatrix& R21 = FEManager.getLifting21(); 

  Eigen::Matrix<double, InputData::n_dofs, InputData::n_dofs, Eigen::RowMajor> ffMatrix;
  Eigen::Matrix<double, InputData::n_dofs, InputData::n_dofs, Eigen::RowMajor> bbMatrix;
  Eigen::Matrix<double, InputData::n_dofs, InputData::n_dofs, Eigen::RowMajor> fbMatrix;
  Eigen::Matrix<double, InputData::n_dofs, InputData::n_dofs, Eigen::RowMajor> bfMatrix;
  Eigen::Matrix<double, InputData::n_dofs, 1> fShapeDerivNormal;
  Eigen::Matrix<double, InputData::n_dofs, 1> bShapeDerivNormal;

  for (SpaMesh::ElementIterator elemIt(mesh); elemIt; ++elemIt) 
  {
    int elem = elemIt.value();
    for (SpaMesh::NeighborIterator neighborIt(mesh,elemIt); neighborIt; ++neighborIt)
    {     
      int neighbor = neighborIt.value();
      if (elem < neighbor)
      {
        FEManager.initInternalFaceFE(mesh, elem, neighbor, neighborIt.index(), InputData::p_order, InputData::q_order);
        FEManager.computeInternalFaceLifting(); 
        ffMatrix.setZero();
        bbMatrix.setZero();
        fbMatrix.setZero();
        bfMatrix.setZero();
        for (unsigned int d = 0; d < InputData::dim; d++)
        {
          bbMatrix.noalias() += InputData::penalty * ((R11(d).transpose() * R11(d)) + (R21(d).transpose() * R21(d))); 
          ffMatrix.noalias() += InputData::penalty * ((R12(d).transpose() * R12(d)) + (R22(d).transpose() * R22(d))); 
          bfMatrix.noalias() += InputData::penalty * ((R11(d).transpose() * R12(d)) + (R21(d).transpose() * R22(d))); 
          fbMatrix.noalias() += InputData::penalty * ((R12(d).transpose() * R11(d)) + (R22(d).transpose() * R21(d))); 
        }
        for (unsigned int j = 0; j < bShapeMatrix.cols(); j++)
        {
          fShapeDerivNormal.setZero();
          bShapeDerivNormal.setZero();
          for (unsigned int d = 0; d < InputData::dim; d++)
          {
            fShapeDerivNormal.noalias() += fShapeDerivMatrix(d).col(j) * normalVec(j)(d);
            bShapeDerivNormal.noalias() += bShapeDerivMatrix(d).col(j) * normalVec(j)(d);
          }
          bbMatrix.noalias() -= 0.5 * jacXweightVecSide[j] * (bShapeMatrix.col(j) * bShapeDerivNormal.transpose());
          bbMatrix.noalias() -= 0.5 * jacXweightVecSide[j] * (bShapeDerivNormal *  (bShapeMatrix.col(j)).transpose());
          ffMatrix.noalias() += 0.5 * jacXweightVecSide[j] * (fShapeMatrix.col(j) * fShapeDerivNormal.transpose());
          ffMatrix.noalias() += 0.5 * jacXweightVecSide[j] * (fShapeDerivNormal *  (fShapeMatrix.col(j)).transpose());
          bfMatrix.noalias() -= 0.5 * jacXweightVecSide[j] * (bShapeMatrix.col(j) * fShapeDerivNormal.transpose());
          bfMatrix.noalias() += 0.5 * jacXweightVecSide[j] * (bShapeDerivNormal *  (fShapeMatrix.col(j)).transpose());
          fbMatrix.noalias() += 0.5 * jacXweightVecSide[j] * (fShapeMatrix.col(j) * bShapeDerivNormal.transpose());
          fbMatrix.noalias() -= 0.5 * jacXweightVecSide[j] * (fShapeDerivNormal *  (bShapeMatrix.col(j)).transpose());
        }
        matrix.add(elem,     elem,     bbMatrix);
        matrix.add(neighbor, neighbor, ffMatrix);
        matrix.add(elem,     neighbor, bfMatrix);
        matrix.add(neighbor, elem,     fbMatrix);
      }
    }
  }
}
   
   
   
