// SpaFEDTe, a Template based C++ library for creating 
// Discontinuous Finite Element Spaces,
// Copyright (C) 2012 Lorenzo Alessio Botti

/* This library is free software; you can redistribute it and/or */
/* modify it under the terms of the GNU Lesser General Public */ 
/* License as published by the Free Software Foundation either */ 
/* version 3.0 of the License, or (at your option) any later version. */

/* This software is distributed in the hope that it will be useful, */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/* Lesser General Public License for more details. */

/* You should have received a copy of the GNU Lesser General Public */
/* License along with this software; if not, a copy of the full */
/* GNU Lesser General Public License can be found at */
/* http://www.gnu.org/licenses/. */

#include <iostream>
#include <assert.h>
#include <time.h>
#include <future>
#include <chrono>

#include "CommonTemplateTypes.hpp"
#include "InputData.hpp"
#include "Mesh.hpp"
#include "ElementIterator.hpp"
#include "ReadMeshVtu.hpp"
#include "LapAssembler.hpp"
#include "ExactSolution.hpp"
#include "VTKWriterSpaMesh.hpp"
#include "Timer.hpp"

#include <Eigen/IterativeLinearSolvers>
#include <Eigen/Core>
#include <Eigen/Sparse>

typedef CommonTemplateTypes::GlobalMatrixType GlobalMatrixType;
typedef CommonTemplateTypes::GlobalVectorType GlobalVectorType;
typedef CommonTemplateTypes::FEManagerType FEManagerType;
typedef CommonTemplateTypes::LiftingOperatorType LiftingOperatorType;
typedef CommonTemplateTypes::MeshType MeshType;

int main( int argc, char* argv[] )
{
  FELib::FELibInit init(argc, argv);
  Eigen::initParallel();
  Timer tmr;
  std::vector<double> l2ErrorNorms;
  std::vector<double> assemblyTimes;
  std::vector<double> solutionTimes;
  std::vector<long unsigned int> matricesNonZeroes;
  std::vector<long unsigned int> nDOFs;

  for (std::string meshFile : InputData::meshSequence)
  {
    MeshType mesh;
    ReadMeshVtu vtuReader;
    //vtuReader.load_mesh("../../meshes_vtu/square_01/square_01_512.vtu", mesh);
    vtuReader.load_mesh(meshFile.c_str(), mesh);
    bool computeFaces = true;
    mesh.computeElemNeighbors(computeFaces);
    mesh.printInfo();
    std::cout<<"Polynomial degree is "<<InputData::p_order<<", DOFs per element "<<InputData::n_dofs<<std::endl;
    
    DofMap dofMap;
    FEManagerType FEManager;
    dofMap.setPOrder(InputData::p_order);
    dofMap.setNVars(1);
    dofMap.computeMatrixSparsity(mesh);

    GlobalMatrixType matrix(dofMap.getGlobalElementNumber()*InputData::n_dofs, dofMap.getGlobalElementNumber()*InputData::n_dofs);
    Eigen::VectorXi matrixNonZeroes(dofMap.getGlobalElementNumber()*InputData::n_dofs);
    for (unsigned int i = 0; i < dofMap.getGlobalElementNumber(); i++)
      matrixNonZeroes.segment<InputData::n_dofs>(i*InputData::n_dofs).noalias() = Eigen::VectorXi::Constant(InputData::n_dofs, dofMap.getLocalNonZeroBlocks()[i]*InputData::n_dofs);
    matrix.reserve(matrixNonZeroes);
    GlobalVectorType rhs(dofMap.getGlobalElementNumber()*InputData::n_dofs);
    rhs.setZero();
    GlobalVectorType solution(dofMap.getGlobalElementNumber()*InputData::n_dofs);
    matricesNonZeroes.push_back(matrixNonZeroes.sum());
    nDOFs.push_back(dofMap.getGlobalElementNumber()*InputData::n_dofs);

    tmr.reset();

    LapAssembler::assemble_volume_contribution(mesh, matrix, rhs, FEManager);
    LapAssembler::assemble_boundary_face_contribution(mesh, matrix, rhs, FEManager);
    LapAssembler::assemble_internal_face_contribution(mesh, matrix, rhs, FEManager);
    double assembly_time = tmr.elapsed();
    if (FELib::processor_id() == 0)
      std::cout<<"rank "<<FELib::processor_id()<<" assembly time "<<assembly_time<<std::endl;
    assemblyTimes.push_back(assembly_time);

    tmr.reset();
    Eigen::ConjugateGradient<GlobalMatrixType> solver;
    solver.setTolerance(1.e-10);
    solver.compute(matrix);
    solution = solver.solve(rhs);
    std::cout << "Eigen Linear solver: #iterations: " << solver.iterations() << ", estimated error: " << solver.error() << std::endl;

    double solution_time = tmr.elapsed();
    solutionTimes.push_back(solution_time);
    if (FELib::processor_id() == 0)
    {
      std::cout<<"rank "<<FELib::processor_id()<<", Solution time "<<solution_time<<std::endl;
      std::cout<<"total step time "<<assembly_time+solution_time<<std::endl;
    }
    const Eigen::Map<const Eigen::Matrix<double,Eigen::Dynamic,InputData::n_dofs,InputData::MajorOption(InputData::p_order)> > 
                     solutionBlocked(solution.data(), mesh.elemCard(), InputData::n_dofs); 
    {
      double l2_error = 0;
      const FEManagerType::ShapeMatrixType& shapeMatrix           = FEManager.getElemShapeFunction();
      const FEManagerType::ElemJacobianVectorType& jacXweightVec  = FEManager.getElemJacobianWeightVec();
      const FEManagerType::ElemQpointVectorType& qpXVec           = FEManager.getElemQpointX();
      for (SpaMesh::ElementIterator elemIt(mesh); elemIt; ++elemIt)
      {
        int elem = elemIt.value();
        FEManager.initVolumeFE(mesh, elemIt, InputData::p_order, InputData::q_order, 2);
        double l2_elem_error = 0;
        for (unsigned int j = 0; j < shapeMatrix.cols(); j++)
        {
          l2_elem_error += pow((solutionBlocked.row(elem) * shapeMatrix.col(j))-InputData::ExactSolution::exact_solution(qpXVec.row(j)),2) * jacXweightVec[j];
        }
        l2_error += l2_elem_error;
      }
      double allL2_error = sqrt(l2_error);
      l2ErrorNorms.push_back(allL2_error);
      if (FELib::processor_id() == 0)
      {
        std::cout<<"dofs "<<dofMap.getGlobalElementNumber()*InputData::n_dofs<<", Error in L2 norm "<<allL2_error<<std::endl;
        std::cout<<std::endl;
      }
      VTKWriterSpaMesh<InputData::dim, FEManagerType> vtkWriter;
      vtkWriter.writeOnTheFly(mesh,FEManager.getFEDataAccessor(),dofMap,solutionBlocked);
    }
  }
  std::cout<<"DOFs, "<<"Matrix nonZeroes, "<<"Error, "<<"Assembly Time, "<<"Solution Time"<<std::endl;
  for (unsigned int i = 0; i < l2ErrorNorms.size(); i++)
    std::cout<<nDOFs[i]<<","<<matricesNonZeroes[i]<<","<<l2ErrorNorms[i]<<","<<assemblyTimes[i]<<","<<solutionTimes[i]<<std::endl;
  for (unsigned int i = 1; i < l2ErrorNorms.size(); i++)
    std::cout<<"Convergence rate (Error in L2 norm) "<<log(l2ErrorNorms[i-1]/l2ErrorNorms[i])/log(2.)<<std::endl;

  return EXIT_SUCCESS;
}

